"""This script is the main entry point for the `axii` CLI."""

import os
import sys

import rich_click as click

# Try to only import lightweight imports here.
from armarx_setup import __version__
from armarx_setup.cli import common, workspace, module, database, self, config, edit, ci
from armarx_setup.core import error
from armarx_setup.core.config import config as axii_config
from armarx_setup.core.util.system import set_process_priorities


click.rich_click.COMMAND_GROUPS = {
    "axii": [
        {
            "name": "Main commands",
            "commands": ["workspace", "module", "database"]
        },
        {
            "name": "Axii",
            "commands": ["self", "config", "edit", "ci"]
        }
    ],

    "axii workspace": workspace.COMMAND_GROUPS,
    "axii module": module.COMMAND_GROUPS,
    "axii database": database.COMMAND_GROUPS,

    "axii self": self.COMMAND_GROUPS,
    "axii config": config.COMMAND_GROUPS,
    "axii edit": edit.COMMAND_GROUPS,
    "axii ci": ci.COMMAND_GROUPS
}


@click.group(**common.group_kwargs)
@click.version_option(__version__)
@click.option("--disable-prompts", default=False, is_flag=True, help="Disables all prompts.")
def entry_point(disable_prompts):
    """Axii: The ArmarX Integrated Installer for setting up and maintaining ArmarX and its
    dependencies.
    """

    common.DISABLE_PROMPTS = os.environ.get("_AXII_DISABLE_PROMPTS", disable_prompts)

    set_process_priorities(nice=10, ionice_class="idle")

    axii_config.console = common.RichConsole()  # Set Rich console as Axii's default console.
    axii_config.executor = common.decorated_execute  # Decorated command executor with Rich output.


entry_point.add_command(workspace.group)
entry_point.add_command(module.group)
entry_point.add_command(database.group)
entry_point.add_command(self.group)
entry_point.add_command(config.group)
entry_point.add_command(edit.group)
entry_point.add_command(ci.group)

if __name__ == '__main__':
    try:
        try:
            return_code = entry_point.main(prog_name="axii", standalone_mode=False)
        except (KeyboardInterrupt, click.Abort) as e:
            raise error.UserAborted() from e
    except error.AxiiError as e:
        error.handle_exception()
        return_code = e.axii_return_code
        error.print_exception_message(exception=e)
    except Exception as e:
        error.handle_exception()
        return_code = error.UNCLASSIFIED_ERROR
        error.print_exception_message(exception=e)

    sys.exit(return_code)
