"""Supplies all subcommands available via `axii ci`."""

import rich_click as click

from armarx_setup.cli import common, complete
from armarx_setup.cli.ci import integration
from armarx_setup.core import error

COMMAND_GROUPS = []  # Default.


@click.group("ci", **common.group_kwargs)
def ci():
    """Provide scriptable services to the CI."""


@ci.group("workspace", **common.group_kwargs)
def ci_workspace():
    """Interact with the active workspace."""


@ci_workspace.command("create", **common.command_kwargs)
@click.option("--ci-module-name",
              help="Name of the CI module.")
@click.option("--ci-module-path",
              help='Path of the already existing CI module, e.g. "${CI_PROJECT_DIR}".')
@click.option("--workspace-name", default="ci_workspace",
              help="Name of the created workspace. Can be used afterwards to activate the workspace.")
@click.option("--workspace-config-path", default=None,
              help="Use the given workspace configuration file.")
@click.option("--with-module", multiple=True,
              help="Add the given module to the workspace.")
@click.option("--ccache/--no-ccache", default=False,
              help="Enable/disable Ccache.")
def ci_workspace_create(**kwargs):
    """Create a workspace for usage in a CI pipeline.

    The _CI module_ is the module name of the project the CI pipeline originates from.
    """

    common.bootstrap(ensure_installation=True, print_logo=False)

    kwargs["with_modules"] = kwargs.pop("with_module")

    try:
        integration.ci_workspace_create(**kwargs)
    except:
        raise


@ci_workspace.command("list-apt-packages", **common.command_kwargs)
@click.option("--all", "-a", is_flag=True,
              help="List all required packages.")
@click.option("--missing", "-m", is_flag=True,
              help="List only missing packages.")
def ci_workspace_list_apt_packages(**kwargs):
    """Print the apt packages required by the workspace as a whitespace-separated list."""

    common.bootstrap(ensure_installation=True, print_logo=False)

    try:
        integration.ci_workspace_list_apt_packages(**kwargs)
    except error.NoWorkspaceActive as e:
        error.handle_exception()
        integration.no_workspace_active(e, action="list required apt packages")
