"""Integrates ci CLI commands to Axii's core."""

import os

import typing as ty

from armarx_setup import console
from armarx_setup.cli import common


def ci_workspace_create(
        ci_module_name: str,
        ci_module_path: str,
        workspace_name: str,
        workspace_config_path: ty.Optional[str],
        with_modules: ty.Tuple[str],
        ccache: bool = False
):
    """Create a workspace and link a module in an active CI job."""

    from armarx_setup.cli.workspace.integration import create_workspace

    workspace_path = os.path.join(os.path.expanduser("~"), workspace_name)

    def workspace_default_config_factory():
        from armarx_setup.core.workspace.config import Config

        modules = dict()
        if ccache:
            modules["tools/ccache/default"] = {}

        cmake_definitions = {
            "CMAKE_BUILD_TYPE": "RelWithDebInfo",
        }
        if ccache:
            cmake_definitions["CMAKE_C_COMPILER_LAUNCHER"] = "$CCACHE"
            cmake_definitions["CMAKE_CXX_COMPILER_LAUNCHER"] = "$CCACHE"

        global_ = {
            "prepare": {
                "cmake": {
                    "definitions": cmake_definitions,
                }
            }
        }
        return Config(
            modules=modules,
            global_=global_,
        )

    # Create workspace.
    ws = create_workspace(
        workspace_path=workspace_path,
        workspace_name=workspace_name,
        workspace_config_path=workspace_config_path,
        workspace_default_config_factory=workspace_default_config_factory,
    )

    # Add requested modules.
    for module in [ci_module_name] + list(with_modules):
        ws.add_module(module, store=False)

    ws.store_config()

    # Create symlink to existing CI module checkout.
    module = ws.module_graph.get_module_with_name(ci_module_name)
    ci_module_target_path = module.path
    os.makedirs(os.path.dirname(ci_module_target_path), exist_ok=True)
    os.symlink(ci_module_path, ci_module_target_path)


def ci_workspace_list_apt_packages(
        all: bool,
        missing: bool,
):
    """Get and print the required apt packages in a whitespace-separated list."""

    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace()
    packages = ws.get_required_apt_packages(all=all, missing=missing)
    console.print(" ".join(packages))


def no_workspace_active(e, action=None):
    from armarx_setup.cli.workspace.integration import no_workspace_active
    no_workspace_active(e, action)
