"""Integrates config CLI commands to Axii's core."""

import sys
import typing as ty

import rich_click as click

from armarx_setup import console
from armarx_setup.core.config import config as axii_config, BuiltInConfigVariables, Config
from armarx_setup.cli import common, complete


def list_config_options(ctx, param, incomplete: str) -> ty.List[str]:
    """Provides shell completion for built-in configuration variables."""

    return complete.filter_completion(BuiltInConfigVariables.get_all_names(), incomplete)


def list_config_option_values(ctx, param, incomplete: str) -> ty.List[str]:
    """Provides shell completion for a built-in configuration variable's values."""

    variable_name = ctx.params["variable_name"]
    possible_values = BuiltInConfigVariables.get_by_name(variable_name).get_allowed_values()
    return complete.filter_completion(possible_values, incomplete)


def config(variable_name, variable_value=None, is_global=True, unset=False, subconfig=False):
    # A workspace must be active if global is true.
    if not is_global:
        from armarx_setup.core.workspace import Workspace
        from armarx_setup.core.error import NoWorkspaceActive
        if not Workspace.is_workspace_active():
            raise NoWorkspaceActive()

    if variable_value is None:
        # Unset variable if --unset was set.
        if unset:
            axii_config.unset(variable_name, is_global=is_global)
            console.print(f"Unset '{variable_name}' ({'global' if is_global else 'workspace-specific'}).")
            axii_config.store_config()
        # If --subconfig was set, then the intention is to create a new subconfig.
        elif subconfig:
            axii_config.set(variable_name, Config(), is_global=is_global)
            console.print(f"Create new subconfig '{variable_name}' ({'global' if is_global else 'workspace-specific'}).")
            axii_config.store_config()
        # Otherwise print it if found.
        else:
            value = axii_config.get(variable_name, is_global=is_global)
            console.print(f"{value}")
    else:
        if unset:
            raise click.UsageError("Only supply one variable name with --unset")

        axii_config.set(variable_name, variable_value, is_global=is_global)
        variable_value_raw = axii_config.get(variable_name, interpret=False, is_global=is_global)
        variable_value = axii_config.get(variable_name, interpret=True, is_global=is_global)
        resolved = f" ({variable_value})" if variable_value_raw != variable_value else ""
        console.print(f"Set '{variable_name}' ({'global' if is_global else 'workspace-specific'}) to "
                      f"{variable_value_raw}{resolved}.")
        axii_config.store_config()


def no_such_config_variable(e, is_global=False):
    console.print(e)
    console.print()
    console.print("You can set it as follows:")
    global_flag = "--global " if is_global else ""
    common.print_commands([
        f"axii config {global_flag}{e.variable_name} \"<new_value>\""
    ])
    sys.exit(1)
