"""Integrates self CLI commands to Axii's core."""

import os
import typing as ty

import armarx_setup
from armarx_setup import console
from armarx_setup.cli import common
from armarx_setup.core import error
from armarx_setup.core.config import config as axii_config
from armarx_setup.self import install as self_install


def install(auto_detect=True, zsh=False, bash=False, yes=False):
    shell_list = self_install.shell_list(auto_detect=auto_detect, zsh=zsh, bash=bash)

    for shell in shell_list:
        install_axii_script_path = self_install.get_install_axii_script_path()
        rc_file_path = self_install.get_rc_file_path(shell)
        console.print(f"Install Axii for {shell} into '{rc_file_path}'.")

        try:
            install_for_shell(shell, rc_file_path, install_axii_script_path, yes=yes)
        except error.NoShellRunComFile as e:
            error.handle_exception()
            console.print(f"[red]Error[/]: {e}")


def install_for_shell(shell, rc_file_path, install_axii_script_path, yes=False):
    if not self_install.is_snippet_installed(shell, rc_file_path, install_axii_script_path):
        installation_snippet = self_install.get_installation_snippet(install_axii_script_path)

        console.print(f"This will append the following lines to '{rc_file_path}':")
        common.print_commands(installation_snippet[1:-1])
        yes = yes or common.input_confirm("Continue?", default_option=False)
        if yes:
            console.print(f"Append snippet to '{rc_file_path}'.")

            self_install.install_snippet_into(shell, rc_file_path, installation_snippet)

            console.print(f"Installed Axii for {shell}.")
            console.print()
            console.print(f"[yellow]Important[/]: For the changes to take effect, please close and "
                          f"re-open all terminals. Alternatively, re-source your '{rc_file_path}' "
                          f"in each of them:")
            common.print_commands([
                f"source {rc_file_path}"
            ])
    else:
        console.print(f"Axii for {shell} already installed in '{rc_file_path}'.")
        console.print()


def update(**kwargs):
    from armarx_setup import self

    self.update(**kwargs)


def change_release(release_channel):
    from armarx_setup import self
    from armarx_setup.cli.common import print_logo

    self.change_release(release_channel=release_channel)

    # Print logo here after branch is switched.
    print_logo()

    console.print(f"Successfully switched to release channel '{release_channel}'.")


def report():
    from armarx_setup import PACKAGE_ROOT as axii_root, __version__ as axii_version
    from armarx_setup.core.util.commands import execute
    from armarx_setup.core.git import repo_status

    axii_repo_status = repo_status(axii_root, fetch=False)

    installed_wheels: ty.Dict[str, str] = {}

    def new_wheel(line):
        nonlocal installed_wheels
        line = line.strip()
        wheel, version = line.split("==")
        installed_wheels[wheel] = version

    def find_venv() -> str:
        from armarx_setup import PACKAGE_ROOT

        os_codename = axii_config.get_os_codename()

        for directory in [f".venv_{os_codename}", ".venv", "venv"]:
            venv = os.path.join(PACKAGE_ROOT, directory)
            if os.path.isdir(venv):
                return directory

        raise FileNotFoundError("No virtual environment found.")

    axii_venv = find_venv()

    try:
        lsb_release = execute("lsb_release --description --short")
    except error.CommandFailed:
        lsb_release = None
    try:
        uname = execute("uname --kernel-name --kernel-release --kernel-version --machine "
                        "--operating-system")
    except error.CommandFailed:
        uname = None
    shell = os.environ.get("SHELL", None)
    try:
        python_path = os.path.join(axii_venv, "bin", "python")
        python = execute(f"{python_path} --version", cwd=axii_root)
    except error.CommandFailed:
        python = None
    try:
        pip_path = os.path.join(axii_venv, "bin", "pip")
        execute(f"{pip_path} freeze", cwd=axii_root, stdout_callback=new_wheel)
    except error.CommandFailed:
        wheels_summary = None
    else:
        wheels_summary = ", ".join(f"{wheel}=={version}" for wheel, version in
                                   installed_wheels.items())

    from rich.table import Table

    shell_integration = "active" if "_axii_installed" in os.environ else "inactive"

    if axii_config.is_axii_locally_installed():
        workspace = f"local in '{axii_config.get_active_workspace_name()}'"
    else:
        workspace = "global"

    table = Table("Scope", "Info", width=80)
    table.add_row("Axii", f"Version: {axii_version}\n"
                          f"Release: {axii_repo_status.active_branch}-{axii_repo_status.commits_behind}\n"
                          f"Install: {workspace}\n"
                          f"Shell integration: {shell_integration}")
    table.add_row("OS", lsb_release)
    table.add_row("Kernel", uname)
    table.add_row("Shell", ":warning:" if shell is None else shell)
    table.add_row("Python", python)
    table.add_row("Pip", wheels_summary)

    console.print(table)


class ReleaseWizard:

    def __init__(self):
        from armarx_setup.core.module.axii_module import axii_remote_repo
        from armarx_setup import \
            __version__ as axii_version, \
            RELEASE_CHANNEL_STABLE as AXII_STABLE_REF, \
            RELEASE_CHANNEL_UNSTABLE as AXII_UNSTABLE_REF, \
            PACKAGE_ROOT as AXII_ROOT

        self.axii_root = AXII_ROOT
        self.axii_remote_repo = axii_remote_repo
        self.axii_version = axii_version
        self.is_major_release = self.axii_version.endswith(".0")
        self.axii_stable_ref = AXII_STABLE_REF
        self.axii_unstable_ref = AXII_UNSTABLE_REF

        self.api = None
        self.project = None
        self.milestone = None

        self.current_maintainer = None
        self.co_maintainers = []
        self.contributors = []
        self.merge_requests = []
        self.release_notification = None

        # Test mode will circumvent security measures and not commit anything on the local repo,
        # nor will it push something or access the GitLab API for anything other than reading.
        self.test_mode = False

    def run(self):
        self._ensure_clean_default_repo()
        self._connect_to_gitlab()
        self._fetch_project_info_from_gitlab()
        self._fetch_maintainers()
        self._local_version_bump()
        self._fetch_current_milestone()
        self._local_changelog_update()
        self._commit_changelog_and_version_bump()
        self._ensure_clean_default_repo()  # Ensure clean state again.
        self._merge_unstable_to_stable_and_push()
        self._create_release()
        if self.is_major_release:
            self._prepare_release_notification()
            self._open_mail()
            self._copy_to_clipboard()

    def _ensure_clean_default_repo(self):
        from armarx_setup.core.git import repo_status

        if self.test_mode:
            return

        axii_repo_status = repo_status(self.axii_root, fetch=False)
        if axii_repo_status.dirty or not axii_repo_status.on_default:
            raise Exception("Repo must be on default branch and clean")

    def _connect_to_gitlab(self):
        import gitlab

        pat = axii_config.get("axii_gitlab_pat")
        if not pat:
            raise Exception("No PAT set")

        console.print(f"Connecting to GitLab instance on "
                      f"'{self.axii_remote_repo.instance_url}' ...")

        self.api = gitlab.Gitlab(url=f"https://{self.axii_remote_repo.instance_url}",
                                 private_token=pat)
        self.api.auth()
        self.current_maintainer = self.api.user

        console.print(f"Connected successfully.  Welcome '{self.current_maintainer.name}'.")

    def _fetch_project_info_from_gitlab(self):
        console.print("Fetching project info ...")
        self.project = self.api.projects.get(self.axii_remote_repo.slug, lazy=True)

    def _fetch_maintainers(self):
        import gitlab

        console.print("Looking up maintainers ...")

        found_current_maintainer = False
        for user in self.project.members.list():
            if user.access_level == gitlab.const.OWNER_ACCESS:
                if user.name != self.current_maintainer.name:
                    user_data = self.api.users.get(user.id)
                    self.co_maintainers.append(user_data)
                else:
                    found_current_maintainer = True
        maintainers_str = [f"'{m.name}'" for m in self.co_maintainers]
        console.print(f"Found co-maintainers: {', '.join(maintainers_str)}")

        if not found_current_maintainer:
            raise Exception("You are not a maintainer of Axii")

    def _local_version_bump(self):
        from importlib import reload

        if not common.input_confirm("Would you like to bump the version now? (Ctrl+C to abort)",
                                    default_option=True):
            return

        axii_init = os.path.join("armarx_setup", "__init__.py")
        common.open_file_in_editor(os.path.join(self.axii_root, axii_init),
                                   kind="top-level Axii module",
                                   tmp_file_name=axii_init)
        reload(armarx_setup)
        self.axii_version = armarx_setup.__version__
        self.is_major_release = self.axii_version.endswith(".0")

    def _fetch_current_milestone(self):
        if self.is_major_release:
            console.print(f"Searching for version milestone for current version "
                          f"'{self.axii_version}' ...")
            milestones = self.project.milestones.list(search=self.axii_version)

            if len(milestones) == 0:
                raise Exception("Could not find milestone for own version")
            if len(milestones) > 1:
                raise Exception("Found more than one milestone matching own version")

            self.milestone = milestones[0]

            if self.milestone.state == "closed":
                self._report_error(f"Found milestone '{self.milestone.title}', but it is already "
                                   f"closed")

            console.print(f"Found milestone with title '{self.milestone.title}'.")

            console.print("Searching contributors to this milestone ...")

            def collect_contributors(assignee):
                if assignee is None:
                    return
                if assignee["name"] == self.current_maintainer.name:
                    return  # Don't thank yourself.
                if assignee["name"] in [m.name for m in self.co_maintainers]:
                    return  # Don't thank other maintainers.
                if assignee["name"] in self.contributors:
                    return  # Only thank once.
                self.contributors.append(assignee["name"])

            for issue in self.milestone.issues():
                collect_contributors(issue.assignee)

            for merge_request in self.milestone.merge_requests():
                self.merge_requests.append(merge_request)
                collect_contributors(merge_request.assignee)

            console.print(f"Contributors are: {', '.join(self.contributors)}.")

        releases = self.project.releases.list()
        for rel in releases:
            if rel.name == f"{self.axii_version}":
                self._report_error("The release does already exist")

    def _local_changelog_update(self):
        console.print("Compiling changelog now ...")

        feature_label = "type::feature"
        docs_label = "type::docs"
        module_label = "type::module"
        module_new_flag_label = "new-module"
        fix_label = "type::fix"
        other_label = "type::chore"

        changelog_section_headings = {
            "features": "**Features:**",
            "modules": "**Modules:**",
            "fixes": "**Fixes:**",
            "other": "_Other:_"
        }

        new_module_flag_hr = "**New**"

        changelog_file_path = os.path.join(self.axii_root, "CHANGELOG.md")

        major, minor, _, _ = self.axii_version.split(".")
        headline_h2 = f"## Releases {major}.{minor}"
        headline_h3 = f"### Release [{self.axii_version}]" \
                      f"({self.axii_remote_repo.releases_url(self.axii_version)})"

        with open(changelog_file_path, encoding="utf-8") as file:
            changelog = file.read()

        if headline_h3 not in changelog:
            search = "<!-- changelog-placeholder -->"
            replace = search
            if headline_h2 in changelog:
                search = headline_h2
                replace = headline_h2
            else:
                replace += "\n\n\n"
                replace += headline_h2

            replace += "\n\n\n"
            replace += headline_h3

            changelog_sections = {
                "features": [],
                "modules": [],
                "fixes": [],
                "other": [],
            }

            def create_changelog_entry(title, author, mr_id, new=False):
                new_text = ""
                if new:
                    new_text = new_module_flag_hr + " "

                entry = f"- {new_text}{title} ("
                if author is not None:
                    entry += f"@{author}, "
                entry += f"!{mr_id})"

                return entry

            for merge_request in self.merge_requests:
                if merge_request.state == "merged":
                    mr_title = merge_request.title
                    author = None
                    mr_id = merge_request.iid
                    if merge_request.assignee is not None:
                        author = merge_request.assignee['username']

                    mr_str = create_changelog_entry(
                        title=mr_title,
                        author=author,
                        mr_id=mr_id
                    )

                    if feature_label in merge_request.labels:
                        changelog_sections["features"].append(mr_str)
                    elif module_label in merge_request.labels:
                        new = False
                        if module_new_flag_label in merge_request.labels:
                            new = True

                        module_names, description = mr_title.split(":", maxsplit=1)
                        module_names = module_names.split(",")
                        module_names = [f"`{m.strip()}`" for m in module_names]

                        new_title = f"{', '.join(module_names)}: {description.strip()}"

                        mr_str = create_changelog_entry(
                            title=new_title,
                            author=author,
                            mr_id=mr_id,
                            new=new
                        )

                        changelog_sections["modules"].append(mr_str)
                    elif fix_label in merge_request.labels:
                        changelog_sections["fixes"].append(mr_str)
                    else:
                        changelog_sections["other"].append(mr_str)
                elif merge_request.state == "closed":
                    pass  # This is fine, as it might have been revamped in a dedicated MR.
                else:
                    raise Exception(f"Merge request {merge_request.iid} is not ready yet")

            changelog_data = [
                "**Update procedure**:",
                "",
                "- `axii self update`",
                "- (with workspace activated) `axii workspace verify`",
                ""
            ]

            for section_id, entries in changelog_sections.items():
                section = changelog_section_headings[section_id]

                if len(entries) > 0:
                    changelog_data += [section, ""] + entries + [""]

            if changelog_data[-1] == "":
                changelog_data.pop(-1)

            console.print("Compiled changelog.")
            console.rule(title=f"Compiled changelog for release {self.axii_version}")
            console.print("\n".join(changelog_data))
            console.rule()
            console.print()

            replace += "\n\n" + "\n".join(changelog_data)

            changelog = changelog.replace(search, replace)

        if not common.input_confirm("Would you like to append the changelog now? (Ctrl+C to abort)",
                                    default_option=True):
            return

        with open(changelog_file_path, "w", encoding="utf-8") as file:
            file.write(changelog)

        common.open_file_in_editor(changelog_file_path,
                                   kind="changelog",
                                   tmp_file_name="CHANGELOG.md")

    def _commit_changelog_and_version_bump(self):
        console.print("The following changes will now be committed\n")
        self._execute("git --no-pager diff", quiet=False)
        if not common.input_confirm("Should these changes be committed and pushed?",
                                    default_option=False):
            raise Exception("Aborted by user")

        if self.test_mode:
            return

        changelog_file_path = os.path.join(self.axii_root, "CHANGELOG.md")
        axii_top_module_file_path = os.path.join("armarx_setup", "__init__.py")
        self._execute(f"git add {changelog_file_path}")
        try:
            self._execute("git commit -m 'chore: Update changelog'")
        except:
            pass
        self._execute(f"git add {axii_top_module_file_path}")
        try:
            self._execute("git commit -m 'chore: Version bump'")
        except:
            pass

    def _merge_unstable_to_stable_and_push(self):
        console.print(f"Pushing changes on '{self.axii_unstable_ref}' ...")
        self._execute("git push")
        self._wait_for_pipeline_to_succeed(self.axii_unstable_ref)

        console.print(f"Merging '{self.axii_unstable_ref}' into '{self.axii_stable_ref}' ...")
        self._execute("git checkout stable")
        self._execute("git pull")
        self._execute("git merge main --no-ff --no-edit")

        console.print(f"Pushing changes on '{self.axii_stable_ref} ...")
        self._execute("git push")
        self._wait_for_pipeline_to_succeed(self.axii_stable_ref)

        console.print(f"Syncing '{self.axii_unstable_ref}' with '{self.axii_stable_ref}' ...")
        self._execute("git checkout main")
        self._execute("git merge stable --no-edit")

        console.print(f"Pushing changes on '{self.axii_unstable_ref}' ...")
        self._execute("git push")
        self._wait_for_pipeline_to_succeed(self.axii_unstable_ref)

        # Testing is currently not in use.  Don't wait for pipeline.
        self._execute("git push origin stable:testing")

    def _create_release(self):
        new_release_dict = {
            "name": f"{self.axii_version}",
            "tag_name": f"{self.axii_version}",
            "tag_message": f"Release {self.axii_version}",
            "ref": self.axii_stable_ref,
            "description": self._get_changelog_for_version(),
            "milestones": [self.milestone.title] if self.milestone is not None else []
        }

        console.print("Creating release ...")
        if not self.test_mode:
            self.project.releases.create(new_release_dict)
            console.print(f"Successfully created release '{self.axii_version}'.")
        else:
            console.print("Error: Test mode.  This would have created a release in non-test-mode.")

    def _prepare_release_notification(self):
        template = """
        Hi everybody,

        We just released Axii version {version}. You can update via `axii self update`.

        For more information and detailed update procedure, see the release notes [1].

        As always, if you experience any problems, feel free to create an issue [2] or contact us.

        {thanks} who contributed by improving Axii or maintaining module definitions.

        Best,
        {signed_by}

        [1] Release notes: {release_notes}
        [2] Create an issue: {create_an_issue}
        """
        template = "\n".join(line.strip() for line in template.strip().split("\n"))

        def first_name(m) -> str:
            return m.name.split(" ")[0]

        signed_by = first_name(self.current_maintainer)
        if self.co_maintainers is not None:
            signed_by = ", ".join(first_name(m) for m in self.co_maintainers) + " and " + signed_by

        thanks = "Thanks to everyone"
        if len(self.contributors) > 0:
            names = ", ".join(self.contributors)
            thanks = f"Thanks especially to {names}, and everyone else"

        self.release_notification = template.format(
            version=self.axii_version,
            release_notes=self.axii_remote_repo.releases_url(self.axii_version),
            create_an_issue=self.axii_remote_repo.new_issue_url(),
            thanks=thanks,
            signed_by=signed_by)

        console.print()

    def _open_mail(self):
        from urllib.parse import quote

        subject = f"Axii Release {self.axii_version}"
        carbon_copy = ",".join(f"{m.name} <{m.public_email}>" for m in self.co_maintainers)
        body = self.release_notification

        console.print(f"[link=mailto:armarx@lists.kit.edu"
                      f"?subject={quote(subject)}"
                      f"&cc={quote(carbon_copy)}"
                      f"&body={quote(body)}]"
                      f"Send release notification mail[/link].")

    def _copy_to_clipboard(self):
        import subprocess
        try:
            command = "xclip -selection clipboard"

            with subprocess.Popen([command], shell=True, stdin=subprocess.PIPE) as process:
                process.communicate(input=self.release_notification.encode())
                console.print("Copied release notification to clipboard.")
        except:
            console.print("Could not copy release notification to clipboard.")

    def _report_error(self, text):
        if self.test_mode:
            console.print(f"Error: {text}.  This would have caused an exception in non-test-mode.")
        else:
            raise Exception(text)

    def _execute(self, command, **kwargs):
        from armarx_setup.core.util.commands import run as exe

        if self.test_mode:
            return

        if "quiet" not in kwargs:
            kwargs["quiet"] = True
        if kwargs["quiet"]:
            kwargs["capture_output"] = True
        exe(command, cwd=self.axii_root, **kwargs)

    def _get_changelog_for_version(self) -> str:
        relevant_section = False
        description = ""

        with open(os.path.join(self.axii_root, "CHANGELOG.md"), encoding="utf-8") as file:
            for line in file:
                if line.startswith("#"):
                    if not relevant_section:
                        if self.axii_version in line:
                            relevant_section = True
                    else:
                        return description.strip()
                if relevant_section:
                    description += line
        return ""

    def _wait_for_pipeline_to_succeed(self, ref):
        import time
        import datetime

        import gitlab

        pipeline_status = "created"
        pid = self.project.id

        working_states = ["created", "waiting_for_resource", "preparing", "pending", "running",
                          "manual", "scheduled"]

        waiting = f"Waiting for pipeline on '{ref}' to succeed ..."

        elapsed = 0

        # Wait times in [seconds].
        initial_wait_time = 5
        update_wait_time = 1

        number_retries = 60  # In case of API HTTP errors, how often to retry before erroring out.

        def wait(wait_time):
            nonlocal elapsed
            time.sleep(wait_time)
            elapsed += wait_time

        with console.status(waiting) as status:
            status.update("Creating pipeline ...")
            wait(initial_wait_time)

            while pipeline_status in working_states:
                wait(update_wait_time)
                try:
                    pipeline = self.api.http_get(f"/projects/{pid}/pipelines/latest", ref=ref)
                except gitlab.exceptions.GitlabHttpError:
                    pipeline_status = "Connection lost, reconnecting ..."
                    number_retries -= 1
                    if number_retries <= 0:
                        raise
                else:
                    pipeline_status = pipeline["status"]
                status.update(f"{waiting} "
                              f"| Current status: {pipeline_status} "
                              f"| Time elapsed: {datetime.timedelta(seconds=elapsed)}")

        if pipeline_status != "success":
            raise Exception("Pipeline failed")

        console.print(f"Pipeline on '{ref}' succeeded.")


def release(**_kwargs):
    release_wizard = ReleaseWizard()
    release_wizard.run()
