"""Interaction with commands and processes."""

import errno
import pty
import multiprocessing as mp
import os
import re
import select
import shlex
import subprocess
import sys
import threading as th
import time
import typing as ty

from rich.rule import Rule

from armarx_setup.core import error
from armarx_setup import console
from armarx_setup.core.util.environ_context import EnvironContext


__all__ = ["PrintCallback", "StdInBuffer", "execute", "run"]


PrintCallback = ty.Callable[[str], None]


class StdInBuffer:
    """Buffer object for stdin that can be used to interact with running processes."""

    def __init__(self, buffer: str = ""):
        self.buffer = b""
        self.lock = mp.Lock()
        if buffer != "":
            self.write_line(buffer)

    def read(self) -> bytes:
        """Reads all available bytes in the buffer.

        Returns
        -------
        str
            The buffer content as bytes.
        """

        with self.lock:
            buffer = self.buffer
            self.buffer = b""
            return buffer

    def is_buffer_empty(self) -> bool:
        """Checks whether the buffer is empty.

        Returns
        -------
        bool
            True if the buffer is empty, else otherwise.
        """

        with self.lock:
            return self.buffer == b""

    def write_line(self, line: str, end: str = "\n", send_eof: bool = False):
        """Writes a new line to stdin.

        By default, a new line character "\n" will be appended to the line if not already the case.

        Parameters
        ----------
        line: str
            The line to be printed.
        end: str
            Line end character which should be appended to a line if not already present.  Set this
            to "" if you don't want to append anything to your input.
        send_eof: bool
            Writes the line and sends an EOF afterwards.  This is equivalent to stdin.write_line
            followed by stdin.write_eof.
        """

        with self.lock:
            if end != "" and not line.endswith(end):
                line += end
            self.buffer += line.encode("utf-8")
        if send_eof:
            self.write_eof()

    def write_eof(self):
        """Writes an EOF character to stdin."""

        with self.lock:
            self.buffer += b"\x04"  # EOF.


class StdInThread:
    """Context manager to manage the StdInBuffer in a thread."""

    def __init__(self, stdin: ty.Optional[StdInBuffer], stdin_fd):
        self.stdin: ty.Optional[StdInBuffer] = stdin
        self.stdin_fd = stdin_fd
        self.thread: ty.Optional[th.Thread] = None
        self.running = th.Event()
        self.running.clear()

    def __enter__(self):
        if self.stdin is not None:
            def run_writer():
                loop_wait = 1e-3  # In seconds.
                self.running.wait()
                while self.running.is_set():
                    if not self.stdin.is_buffer_empty():
                        os.write(self.stdin_fd, self.stdin.read())
                    time.sleep(loop_wait)

            self.thread = th.Thread(target=run_writer)
            self.thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.running.clear()
        if self.stdin is not None:
            self.thread.join()


class OutputStreamBuffers:
    """General management and context manager to handle output streams and buffers thereof."""

    def __init__(
            self,
            stdout_callback: ty.Optional[ty.Callable] = None,
            stderr_callback: ty.Optional[ty.Callable] = None
     ):
        self.stdout: ty.List[str] = []
        self.stderr: ty.List[str] = []
        self.stdout_buffer: bytes = b""
        self.stderr_buffer: bytes = b""

        self.stdout_callback = stdout_callback
        self.stderr_callback = stderr_callback

    def print_stdout(self, text: str):
        self._print(text, self.stdout, self.stdout_callback)

    def print_stderr(self, text: str):
        self._print(text, self.stderr, self.stderr_callback)

    def get_stdout(self) -> str:
        return "\n".join(self.stdout)

    def get_stderr(self) -> str:
        return "\n".join(self.stderr)

    def buffer_or_print(self, data: bytes, stderr: bool = False):
        if not stderr:
            self.buffer_or_print_stdout(data)
        else:
            self.buffer_or_print_stderr(data)

    def buffer_or_print_stdout(self, data: bytes):
        self.stdout_buffer = self._buffer_or_print(data, self.stdout_buffer, self.print_stdout)

    def buffer_or_print_stderr(self, data: bytes):
        self.stderr_buffer = self._buffer_or_print(data, self.stderr_buffer, self.print_stderr)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Flush buffers by emulating a carriage return line feed if they are not empty.
        if self.stdout_buffer != b"":
            self.buffer_or_print_stdout(b"\r\n")
        if self.stderr_buffer != b"":
            self.buffer_or_print_stderr(b"\r\n")

    @staticmethod
    def _print(text: str, text_buffer: ty.List[str], print_fn: ty.Optional[ty.Callable]):
        if print_fn is not None:
            print_fn(text)
        text_buffer.append(text)

    @staticmethod
    def _buffer_or_print(data: bytes, data_buffer: bytes, print_fn: ty.Callable):
        data_buffer += data
        new_line = b"\r\n"

        if data != b"\r":
            for search in [new_line, b"\r"]:
                if search in data_buffer:
                    to_print, data_buffer = data_buffer.split(search, maxsplit=1)
                    line = to_print.decode("utf-8")
                    print_fn(line)
                    break

        return data_buffer


class PseudoTerminals:
    """Context manager handle pseudo-terminals."""

    def __init__(self):
        # Open pseudo-terminals.
        self.masters, self.slaves = zip(pty.openpty(), pty.openpty(), pty.openpty())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_masters()

    def std_streams_to_dict(self):
        return {
            "stdin": self.slaves[0],
            "stdout": self.slaves[1],
            "stderr": self.slaves[2],
        }

    def stdin(self):
        return self.masters[0]

    def close_slaves(self):
        for file_descriptor in self.slaves:
            os.close(file_descriptor)

    def close_masters(self):
        for file_descriptor in self.masters:
            os.close(file_descriptor)

    def get_readables(self):
        return {
            self.masters[1]: sys.stdout.buffer,  # store buffers separately
            self.masters[2]: sys.stderr.buffer,
        }

    def is_stdout(self, file_descriptor):
        return file_descriptor == self.masters[1]

    def is_stderr(self, file_descriptor):
        return file_descriptor == self.masters[2]


class ExecuteErrors:
    """Context manager to handle errors occurring during subprocess execution."""

    def __init__(self, cmd, cwd):
        self.cmd = cmd
        self.cwd = cwd

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is KeyboardInterrupt:
            raise error.UserAborted from exc_val
        # No such file or directory in case cwd does not exist.
        if exc_type is OSError and exc_val.errno == errno.ENOENT:
            raise error.NoSuchCommandCwd(cmd=self.cmd, cwd=self.cwd) from exc_val


def execute(
        cmd: ty.Union[ty.List[str], str],
        cwd: ty.Optional[str] = None,
        stdin: ty.Optional[ty.Union[str, StdInBuffer]] = None,
        stdout_callback: ty.Optional[PrintCallback] = None,
        stderr_callback: ty.Optional[PrintCallback] = None,
        env: ty.Optional[ty.Dict[str, str]] = None
):
    """Executes a command, allowing various forms of interaction with the command's process.

    Parameters
    ----------
    cmd: ty.Union[ty.List[str], str]
        The command to run as list of strings or string.  Each element of that list can be
        considered quoted.  If a string is passed, it will be shell-lexed using shlex.
    cwd: ty.Optional[str]
        The working directory the command should be run in.
    stdin: ty.Optional[ty.Union[str, StdInBuffer]]
        Either a string to be written on stdin after the command has been started, or a StdInBuffer
        object with that both processes can interact.
    stdout_callback: ty.Optional[PrintCallback]
        Called whenever a new line has been written to stdout.
    stderr_callback: ty.Optional[PrintCallback]
        Called whenever a new line has been written to stderr.
    env: ty.Optional[ty.Dict[str, str]]
        Environment variables that should be set when the command is run, and reset once it's
        finished.

    Returns
    -------
    str
        The commands output via stdout.

    Raises
    ------
    error.NoSuchCommandCwd
        If the command cwd does not exist.
    error.CommandFailed
        If the command exited with a non-zero return code.
    error.UserAborted
        If the user aborted command execution (e.g., SIGINT).
    OSError
        If the communication with the operating system fails.
    """

    # If environment is set, create a environment context and run command recursively with no
    # environment being set.
    if env is not None:
        with EnvironContext(env, expand_environment_variables=True):
            return execute(
                cmd=cmd,
                cwd=cwd,
                stdin=stdin,
                stdout_callback=stdout_callback,
                stderr_callback=stderr_callback,
                env=None)

    if isinstance(cmd, str):
        cmd = shlex.split(cmd, posix=True)

    if isinstance(stdin, str):
        stdin = StdInBuffer(stdin)

    output_streams = OutputStreamBuffers(
        stdout_callback=stdout_callback,
        stderr_callback=stderr_callback)

    with    ExecuteErrors(cmd=cmd, cwd=cwd), \
            PseudoTerminals() as pseudo_terminals, \
            subprocess.Popen(cmd, cwd=cwd, **pseudo_terminals.std_streams_to_dict()) as process, \
            output_streams as output_streams, \
            StdInThread(stdin=stdin, stdin_fd=pseudo_terminals.stdin()) as stdin_thread:
        pseudo_terminals.close_slaves()
        readable = pseudo_terminals.get_readables()
        stdin_thread.running.set()

        while readable:
            for file_descriptor in select.select(readable, [], [])[0]:
                data: ty.Optional[bytes] = None
                try:
                    data = os.read(file_descriptor, 1)  # Read one byte.
                except OSError as e:
                    error.handle_exception(
                        expected_because="On some systems, OSError with errno = EIO is thrown to "
                                         "indicate EOF on stdin, stdout or stderr")
                    if e.errno != errno.EIO:
                        raise
                    data = None
                finally:
                    if data:
                        is_stderr: bool = pseudo_terminals.is_stderr(
                            file_descriptor=file_descriptor)
                        output_streams.buffer_or_print(data=data, stderr=is_stderr)
                    else:  # EOF either via exception or other means.
                        del readable[file_descriptor]

    if process.returncode != 0:
        raise error.CommandFailed(
            cmd=cmd,
            cwd=cwd,
            return_code=process.returncode,
            stdout=output_streams.get_stdout(),
            stderr=output_streams.get_stderr())

    return output_streams.get_stdout().rstrip()


def strip_ansi(in_str: str) -> str:
    pattern = re.compile(r'\x1B\[\d+(;\d+){0,2}m')
    return pattern.sub('', in_str)


def run(cmd, cwd=None, capture_output=False, quiet=False, env_ctx: EnvironContext = None):
    in_cwd = "" if cwd is None else f" in '[link=file://{cwd}]{cwd}[/link]'"
    if not quiet:
        in_env_ctx = " "
        if env_ctx is not None:
            in_env_ctx = f" with environment variables {env_ctx.print_env_dict()} "
        console.print(f"Run '{cmd}'{in_cwd}{in_env_ctx}...")
        console.print(Rule(title="Command output"))

    kwargs = dict(cwd=cwd, shell=True)
    if cwd is not None:
        kwargs["cwd"] = cwd
    if capture_output:
        kwargs["stdout"] = subprocess.PIPE
        kwargs["stderr"] = subprocess.PIPE

    result = subprocess.run(cmd, **kwargs)
    if not quiet:
        color = 'green' if result.returncode == 0 else 'red'
        console.print(Rule(title=f"Command exited - Return code "
                                 f"[{color}]{result.returncode}[/{color}]"))

    if result.returncode != 0:
        raise error.CommandFailed(cmd=cmd,
                                  stdout=result.stdout.decode().strip() if capture_output else None,
                                  return_code=result.returncode,
                                  stderr=result.stderr.decode().strip() if capture_output else None)

    if capture_output:
        stdout = result.stdout.decode().strip()
        stderr = result.stderr.decode().strip()
        return f"{stdout}\n{stderr}" if stderr else stdout
