#!/bin/bash

# Default arguments.
call_to_script="$0 $@"
shell="bash"
no_shell=false
apt=false
dev=false
yes=false
environment="production"

# Argument parsing.
while [[ $# -gt 0 ]]; do
  case $1 in
    --no-shell)
      no_shell=true
      shift
      ;;
    --bash)
      shell="bash"
      shift
      ;;
    --zsh)
      shell="zsh"
      shift
      ;;
    --apt)
      apt=true
      shift
      ;;
    --dev)
      dev=true
      shift
      ;;
    --yes)
      yes=true
      shift
      ;;
    --*|-*)
      echo "Unknown option $1"
      exit 1
      ;;
    *)
      #POSITIONAL_ARGS+=("$1") # save positional arg
      shift # past argument
      ;;
  esac
done

if $dev; then
  environment="development"
fi

apt_install_args=""
if $yes; then
  apt_install_args="--yes"
fi

python_bin=""
if which python3 > /dev/null; then
  python_bin="$(which python3)"
elif which python > /dev/null; then
  python_bin="$(which python)"
fi

axii_dir="$( cd -- "$(dirname "$0")" >/dev/null 2>&1 ; pwd -P )"
bin_name=axii
source /etc/os-release
venv_dir="$axii_dir/.venv_$VERSION_CODENAME"
hline="--------------------------------------------------------------------------------------"


####################################################################################################
echo "$hline"
echo "Verifying apt dependencies ..."
echo "$hline"
echo ""

function apt_packages_missing()
{
  echo
  echo "Error: Required apt packages are missing.  If you have root access, re-run the bootstrap " \
       "with '--apt':"
  echo ""
  echo "    $call_to_script --apt"
  echo ""
  echo "If you do not have root access, contact your system administrator."
}

if $apt && [[ -z "$python_bin" ]]; then
  sudo apt-get install $apt_install_args python3
  python_bin="$(which python3)"
fi

if [[ -z "$python_bin" ]]; then
  echo "No apt packages python3 (or python) found."
  apt_packages_missing
  exit 1
else
  echo "Found Python in '$python_bin'."
fi

packages=$($python_bin "$axii_dir/armarx_setup/app/apt_requirements.py" \
                       "$axii_dir/requirements/apt/$environment.txt")
packages_list=$(echo "$packages" | tr '\n' ' ')  # New lines to whitespaces.

if $apt; then
  sudo apt-get install $apt_install_args $packages_list
else
  if ! echo "$packages" | xargs dpkg -V; then
    apt_packages_missing
    sleep 1s
    echo ""
    echo "Trying to continue with missing packages."
  fi
fi


####################################################################################################
echo ""
echo "$hline"
echo "Create virtual environment in $venv_dir"
echo "$hline"
echo ""

$python_bin -m venv "$venv_dir"
source "$venv_dir/bin/activate"
pip install --upgrade pip
pip install --upgrade setuptools
pip install wheel
pip install -r "$axii_dir/requirements/$environment.txt"


####################################################################################################
if $no_shell; then
  echo ""
  echo "$hline"
  echo "Not installing Axii shell integration since '--no-shell' is set.  This is okay if the main"\
       "Axii installation is global, and you are installing a workspace-specific Axii override.  If"\
       "you want to install Axii at a later point, see 'axii self install --help'."
  echo "$hline"
  echo ""
else
  echo ""
  echo "$hline"
  echo "Installing Axii by running 'axii self install --no-auto-detect --$shell'."
  echo "$hline"
  echo ""

  axii_install_args="--no-auto-detect --$shell"
  if $yes; then
    axii_install_args="$axii_install_args --yes"
  fi

  "$axii_dir"/bin/"$bin_name" self install $axii_install_args
fi


####################################################################################################
echo ""
echo "$hline"
echo "Run axii --help"
echo "$hline"
echo ""

"$axii_dir"/bin/"$bin_name" --help
