# Module Authors' Documentation

This page contains information relevant for authors of module definitions and module databases.

[[_TOC_]]

---


## Modules and the Module Graph

Axii is based on **_modules_**.

Let's start with an example:
- There is a module `armarx/ArmarXCore` which clones, configures, builds and installs 
the ArmarX package `ArmarXCore`.
- In addition, the module `armarx/ArmarXCore` depends on, or _requires_, another module named
`simox`, which clones, configures, builds and installs the standalone CMake package `Simox`.
- The module `simox`, in turn, requires the module `apt/eigen3`, which installs 
the linear algebra library `Eigen` via a system package.

This dependency chain can be represented by a directed (acyclic) graph, 
where modules are nodes and dependencies build edges,
```
(armarx/ArmarXCore) --> (simox) --> (apt/eigen3)
```
or, visually:

![Module Graph](resources/module-graph.svg)

This **_module graph_** is the heart of Axii.
In general, modules can clone repositories, build projects, install system packages, and more.
Most importantly, modules have dependencies to other modules (their _required modules_).
They build the module graph, which represents a program of how your workspace should be set up. 

Apropos: Your workspace specifies one or several modules it would like to have installed
in its config file `armarx-workspace.json`.
These modules are added to the module graph, as well as their dependencies, 
and their dependencies, and so on.

You can even visualize the active module graph by running 
(optional arguments in `[]`):
```bash
axii workspace info --graph [--reduce] [--all]
```

The module name, e.g. `armarx/ArmarXCore`, is also the path inside your workspace
where a module representing a code base will be cloned or downloaded to 
(e.g. `~/code/armarx/ArmarXCore` if your workspace is `~/code`).
At the same time, they are paths to `.json` files in the 
[`data/modules`](data/modules) directory,
which specifies what the module does. 

A **module definition** is a `.json` file in a module database.
Conversely, a **module database** is a directory containing module definitions, 
potentially in a nested directory structure.


## Module Databases

The built-in module database is [`data/modules`](/data/modules).
In addition, Axii uses the environment variable `ARMARX_MODULES` to find additional
module databases and the modules in them.
Similar to `PATH` and `PYTHONPATH`, `ARMARX_MODULES` is a colon-separated list of 
module databases (for example, 
`$HOME/my-module-db:$HOME/Projects/my-project/axii-module-db`).

When searching for a module definition, 
Axii checks each database and uses the first matching module definition it finds.
The order is as follows:
1. Databases in `ARMARX_MODULES`, from front to back.
2. The built-in module database `data/modules`.

This allows two use cases especially:
- Add new modules to Axii without changing the built-in database.
- Overwrite module definitions from the built-in database.


## Workspace Upgrade Procedure

When **upgrading** a workspace, the different steps of all modules are executed in a specific order:

![Upgrade Procedure](resources/steps.svg)

The vertical axis (rows) illustrates the modules in order of their flattened dependency chain
(from no dependencies to many dependencies).
The numbers and arrows indicate time.

1. Generate the **environment** (set environment variables, source files). 
   This step does not occur in the module definition, as it is derived from the other steps (e.g. install).  
2. Run the **system** steps for all modules (e.g., install system packages).
3. **Update** all modules (e.g., clone or pull Git repositories, download archives).
4. For each module:
   1. **Prepare** the module (e.g., run `cmake`, create Python virtual environments).
   2. **Build** the module (e.g., run `cmake --build`).
   3. **Install** the module (e.g., run `cmake --install`).
5. Run final **conclude** steps for all modules.

In other words, the main steps **prepare**, **build** and **install** are run _per-module_.
The steps **system** and **update** are run for all modules _in parallel_ to speed up the process.
The **conclude** step is run per-module, but only after all other steps have been run.


## Module Definitions

Module definitions are descriptive `.json` files which define 
how a module is fetched/updated, built, installed, etc, and which other modules it requires.

The full top-level structure of a module definition is as follows (all of these are optional):

```json
{
  "general": {...},
  
  "system": {...},
  "update": {...},
  "prepare": {...},
  "build": {...},
  "install": {...},
  "conclude": {...},
  
  "required_modules": {...},
  
}
```

In the following you will find a detailed description of each part.


### `general`

The `general` section of a module can have general information, such as a project 
URL. It is mainly informative, but is used to create links in the terminal or 
other quality of life information displays.


#### `general.url` (recommended)

The general URL where more information about the project can be found. For example
the GitHub page, GitLab page, website, etc.

#### `general.authors` (recommended) 

List of authors maintaining this module definition (not the project itself!).

Author names are a string in the style of `"Display Name <email@example.com>"`.


#### `general.database`

*TODO*


##### `general.database.for_namespace` (required)

*TODO*


#### `general.deprecated`

If set, the module is considered deprecated and Axii will discourage its usage.
It will still be possible though.


##### `general.deprecated.since` (required)

Date in format `yyyy-mm-dd` of when the module was declared deprecated.


#### `general.deprecated.message` (required)

Deprecation message displayed to the user


#### `general.deprecated.removal`

Date in format `yyyy-mm-dd` as of when the module is planned to be deleted. This
is just an information for the user. If not deleted manually from the module
database, the module will still be available even after this deadline.


#### `general.deprecated.superseded_by`

This property is a dictionary which has the same format as 
[`required_modules`](#required_modules). Axii can use this to generate suggestions
for the user to drop the deprecated modules and use these suggested by the maintainer
instead.

In most cases, a module will be deprecated and superseded by exactly one new module,
but it might happen that what was once a module becomes the feature or another 
module or similar, which can also be reflected with this structure.


### `system`

*TODO*


### `update`

Strategies about how to get the code or resources and how to update them.


#### `update.git`

Use Git to get the code and update it. The repository location can most conveniently
be configured with the corresponding code hoster shorthands which are mutually exclusive:

- For GitLab.com: `gitlab_com_slug`
- For H2T GitLab: `h2t_gitlab_slug`
- For a generic GitLab instance: `gitlab_instance` and `gitlab_slug`
- For GitHub.com: `github_slug`
- Low level generic SSH and HTTPS URLs: `ssh_url` and `https_url`


##### `update.git.gitlab_com_slug`

If the repository is hosted on GitLab.com. The slug of the project on GitLab.com, i.e.,
the part after `https://gitlab.com/` in the URL.

Example:
```json
"update": {
  "git": {
    "gitlab_com_slug": "graphviz/graphviz"
  }
}
```


##### `update.git.h2t_gitlab_slug`

If the repository is hosted on the H2T GitLab instance. The slug of the project on 
the H2T GitLab instance, i.e., the part after `https://git.h2t.iar.kit.edu/` in the URL.

Example:
```json
"update": {
  "git": {
    "h2t_gitlab_slug": "sw/armarx/meta/axii"
  }
}
```


##### `update.git.gitlab_instance`

If the repository is hosted on another GitLab instance. The URL of the GitLab instance, 
i.e., the part after `https://` and before the first `/`.  For example
`git.h2t.iar.kit.edu` for the H2T GitLab instance or `gitlab.com` for GitLab.com.  Must
be used with `gitlab_slug`.

Example:
```json
"update": {
  "git": {
    "gitlab_instance": "git.h2t.iar.kit.edu"
    "gitlab_slug": "sw/armarx/meta/axii"
  }
}
```


##### `update.git.gitlab_slug`

If the repository is hosted on another GitLab instance. The slug of the project on 
that instance, i.e., the part after `https://domain.tld/` in the URL.
Must be used with `gitlab_instance`.

Example:
```json
"update": {
  "git": {
    "gitlab_instance": "git.h2t.iar.kit.edu"
    "gitlab_slug": "sw/armarx/meta/axii"
  }
}
```


##### `update.git.github_slug`

If the repository is hosted on GitHub. The slug of the project on GitHub, i.e., the part 
after `https://github.com/` in the URL.

Example:
```json
"update": {
  "git": {
    "github_slug": "Kitware/CMake"
  }
}
```


##### `update.git.ssh_url`

The full URL to clone the repository via SSH.  Should be in the form of:
`git@domain.tld:slug/repo.git`.

Example:
```json
"update": {
  "git": {
    "ssh_url": "git@git.h2t.iar.kit.edu:sw/armarx/meta/axii.git"
  }
}
```


##### `update.git.https_url`

The full URL to clone the repository via HTTPS.  Should be in the form of:
`https://domain.tld/slug/repo.git`.

Example:
```json
"update": {
  "git": {
    "https_url": "https://git.h2t.iar.kit.edu/sw/armarx/meta/axii.git"
  }
}
```


##### `update.git.default_checkout_ref`

The default reference to check out while cloning.  The reference can be either a branch
or a tag.

Example:
```json
"update": {
  "git": {
    "h2t_gitlab_slug": "sw/armarx/meta/axii",
    "default_checkout_ref": "stable"
  }
}
```


##### `update.git.default_checkout_commit`

The default commit hash to check out after cloning.

Example:
```json
"update": {
  "git": {
    "h2t_gitlab_slug": "sw/armarx/meta/axii",
    "default_checkout_commit": "1234567890abcdef"
  }
}
```


##### `update.git.submodules`

If true, Git submodules are cloned and updated recursively.

Example:
```json
"update": {
  "git": {
    "h2t_gitlab_slug": "sw/armarx/meta/axii",
    "submodules": true
  }
}
```


##### `update.git.lfs`

If true, Git LFS will be enabled for this project. 

Example:
```json
"update": {
  "git": {
    "h2t_gitlab_slug": "sw/armarx/meta/axii",
    "lfs": true
  }
}
```


##### `update.git.ignore_patterns`

List of local ignore patterns. An ignore pattern looks like the entry if a `.gitignore`
file to ignore expected changes in a module not covered by the `.gitignore` file of that
module's repository.

Example:
```json
"update": {
  "git": {
    "h2t_gitlab_slug": "sw/armarx/meta/axii",
    "ignore_patterns": [
      "build",
      ".venv",
      "*.o"
    ] 
  }
}
```


##### `update.git.prefer_https`

If true, the HTTPS URL will be preferred over the SSH URL. 

Example:
```json
"update": {
  "git": {
    "h2t_gitlab_slug": "sw/armarx/meta/axii",
    "prefer_https": true
  }
}
```


#### `update.download`

Download a file from the internet.


##### `update.download.url`

The URL of the file to download.

Example:
```json
"update": {
  "download": {
    "url": "https://example.com/some-file.zip"
  }
}
```


##### `update.download.file` (`: str`)

The name of the downloaded file on the local filesystem after download.  Can be useful when 
downloading form a generic redirect URL.

Example:
```json
"update": {
  "download": {
    "url": "https://example.com/download",
    "file": "some-file.zip"
  }
}
```


##### `update.download.md5_checksum` (`: Optional[str] = null`)

The MD5 checksum of the file for verification.  This checksum will be verified after download.
If the checksum verification fails, the download sub-step will fail.

The checksum can be generated using `md5sum` as follows:

```shell
md5sum <name of the file>
```

Example:
```json
"update": {
  "download": {
    "url": "https://example.com/some-file.zip",
    "md5_checksum": "d41d8cd98f00b204e9800998ecf8427e"
  }
}
```


##### `update.download.make_executable` (`: bool = false`)

If true, the downloaded file will be made executable.

Example:
```json
"update": {
  "download": {
    "url": "https://example.com/some-installer.sh",
    "make_executable": true
  }
}
```


### `prepare`

Strategies about how to prepare the code or the environment.


#### `prepare.cmake`

Use CMake as a strategy to prepare the code for C/C++ projects.


##### `prepare.cmake.project_name`

*TODO*


##### `prepare.cmake.project_dir`

*TODO*


##### `prepare.cmake.definitions`

*TODO*


##### `prepare.cmake.generator`

*TODO*


##### `prepare.cmake.disable_cmake_warnings`

If true, all CMake warnings (dev and deprecation warnings) are disabled and will not be printed.
Recommended for third party dependencies, but should not be set for own projects.

Example:
```json
"prepare": {
  "cmake": {
    "disable_cmake_warnings": true
  }
}
```


##### `prepare.cmake.disable_build_warnings`

If true, all compiler warnings are disabled and will not be printed.
Recommended for third party dependencies, but should not be set for own projects.

Example:
```json
"prepare": {
  "cmake": {
    "disable_build_warnings": true
  }
}
```


##### `prepare.cmake.disable_all_warnings`

Shorthand for setting `disable_cmake_warnings` and `disable_build_warnings` to `true`.

If true, all warnings are disabled and will not be printed.
Recommended for third party dependencies, but should not be set for own projects.

Example:
```json
"prepare": {
  "cmake": {
    "disable_all_warnings": true
  }
}
```


#### `prepare.extract` (`: Optional[Dict] = null`)

Use an archive tool to extract an archive file.

##### `prepare.extract.file` (`: str`)

Path to the archive file to be extracted.

The path may be absolute or relative to the modules root directory in the workspace.

Example:
```json
"prepare": {
  "extract": {
    "file": "some-file.zip"
  }
}
```


##### `prepare.extract.file_ext` (`: Optional[str] = null`)

Explicitly set the file extension of it is not clear from the file name alone.

If unset, the file extension will be inferred from the file name in `prepare.extract.file`.

Possible values: `.tar.gz`, `.tar.xz`, `.tgz`, `.deb`, `.zip`

Example:
```json
"prepare": {
  "extract": {
    "file": "some-file",
    "file_ext": ".zip"
  }
}
```


##### `prepare.extract.extract_into` (`: str | None = none`)

Explicitly set the path to extract the archive into.

If unset, the path will be inferred form the file name in `prepare.extract.file` by stripping the
file extension.

The path may be absolute or relative to the modules root directory in the workspace.

Example:
```json
"prepare": {
  "extract": {
    "file": "some-file.zip",
    "extract_into": "user-defined-folder"
  }
}
```


### `build`

*TODO*


#### `build.cmake`

*TODO*


##### `build.cmake.env` (`: dict[str, str] = {}`)

*TODO*


#### `build.run`

Executes the given command in a pseudo-terminal.


##### `build.run.command` (`: str | list[str] | None = None`)

Command to run. Simple commands can be given as one string (e.g., `cp from_path to_path`).
If the command is more complex and needs escaping, or paths contain spaces etc., it can also be
passed as a list of strings (e.g., `["cp", "from path", "to path"]`).


##### `build.run.comment` (`: str = ""`)

Comment that is printed before actually running the command in order to provide the user with information about what the
command is going to do semantically.

Example:
```json
"build": {
  "run": {
    "command": "cp MODULE_DEFINITION_RESOURCE_DIR/config $MODULE_PATH/config"
    "comment": "Copy configuration file from Axii module database to workspace"
  }
}
```


##### `build.run.run_in_shell` (`: bool = False`)

If set to true, the command is run in a shell (`bash` by default).
This is equivalent of executing the command `bash -c <original command>`.


### `install`

*TODO*


#### `install.env`

*TODO*


#### `install.cmake`

*TODO*


#### `install.source`

*TODO*


#### `install.run`

Executes the given command in a pseudo-terminal.


##### `install.run.command` (`: str | list[str] | None = None`)

Command to run. Simple commands can be given as one string (e.g., `cp from_path to_path`).
If the command is more complex and needs escaping, or paths contain spaces etc., it can also be
passed as a list of strings (e.g., `["cp", "from path", "to path"]`).


##### `install.run.comment` (`: str = ""`)

Comment that is printed before actually running the command in order to provide the user with information about what the
command is going to do semantically.

Example:
```json
"install": {
  "run": {
    "command": "cp MODULE_DEFINITION_RESOURCE_DIR/config $MODULE_PATH/config"
    "comment": "Copy configuration file from Axii module database to workspace"
  }
}
```


##### `install.run.run_in_shell` (`: bool = False`)

If set to true, the command is run in a shell (`bash` by default).
This is equivalent of executing the command `bash -c <original command>`.


#### `install.executables`

*TODO*


### `conclude`

*TODO*


#### `conclude.run`

Executes the given command in a pseudo-terminal.


##### `conclude.run.command` (`: str | list[str] | None = None`)

Command to run. Simple commands can be given as one string (e.g., `cp from_path to_path`).
If the command is more complex and needs escaping, or paths contain spaces etc., it can also be
passed as a list of strings (e.g., `["cp", "from path", "to path"]`).


##### `conclude.run.comment` (`: str = ""`)

Comment that is printed before actually running the command in order to provide the user with information about what the
command is going to do semantically.

Example:
```json
"conclude": {
  "run": {
    "command": "cp MODULE_DEFINITION_RESOURCE_DIR/config $MODULE_PATH/config"
    "comment": "Copy configuration file from Axii module database to workspace"
  }
}
```


##### `conclude.run.run_in_shell` (`: bool = False`)

If set to true, the command is run in a shell (`bash` by default).
This is equivalent of executing the command `bash -c <original command>`.


### `required_modules`

The `required_modules` section specifies which other modules this module requires
(_aka_ depends on) and is used to build the module graph.

Note about transitive dependencies: Only include transitve dependencies if your module has a public dependency to the required module.
So in case A depends on B, B depends on C and A depends on C you should only add C to the `required_modules` of A if...
- the top-level CMakeList in A has a public dependency to C so there is a armarx_find_package(PUBLIC C) in the CMakeList
- in the prepare step of the axii module of A a python package of module C is listed in the "install_editable" command


## General Features in Module Definitions


### Available Environment Variables

While parsing a module definition, or while executing steps of a module, Axii defines several environment variables
which can be used to ease the setup or scripting.

For use in the workspace:

- `MODULE_PATH`: Path to the module inside the current workspace.  
**Example:** Workspace is `/path/to/workspace` and module is `ns/module` => `MODULE_PATH` = `/path/to/workspace/ns/module`.

For use in Axii's module database, for example to access resources (scripts, config files, templates, ...):

- `MODULE_DEFINITION_DIR`: Path to the directory where the module definition file resides.  
**Example:** Module definition is `/path/to/module_database/ns/module.json` => `MODULE_DEFINITION_DIR` = `/path/to/module_database/ns`.
- `MODULE_DEFINITION_RESOURCE_DIR`: Path to the resource directory of the module. The resource directory of a module is a 
*directory* with the same name as the module (except for the `.json` file extension). It is a convention to group 
resources (such as scripts, config files, templates, etc.) belonging to a module.  
**Example:** Module definition is `/path/to/module_database/ns/module.json` => `MODULE_DEFINITION_RESOURCE_DIR` = `/path/to/module_database/ns/module`.


### JSON Comments

Axii treats all dictionary keys and (string) list elements starting with a `#` as comments
and removes them after loading before any further interpretation happens.

For example, the JSON file
```json
{
  "update": {...},
  "build": {...},
  
  "# Not necessary anymore": {...},
  "# install": {...},
  
  "required_modules": {
    "other_module_a": {},
    "# other_module_b": {}
  }
}
```
is interpreted as
```json
{
  "update": {...},
  "build": {...},
  
  "required_modules": {
    "other_module_a": {}
  }
}
```


### OS-Specific Definitions

To cover OS-specific differences, any value in a module definition can be written 
in a switch-case like syntax using the `axii.switch_os` key as follows:

- Dictionary value:
  ```json
  {
    "a": "A",
    "b": {
      "axii.switch_os": {
        "bionic": "ubuntu18-B",
        "focal": "ubuntu20-B",
        "jammy": "ubuntu22-B",
        "buster": "debian10-B",
        "default": "B"
      }
    },
    "c": "C"
  }
  ```
- List element:
  ```json
  [
    "A",
    {
      "axii.switch_os": {
        "bionic": "ubuntu18-B",
        "focal": "ubuntu20-B",
        "jammy": "ubuntu22-B",
        "buster": "debian10-B",
        "default": "B"
      }
    },
    "C"
  ]
  ```

The case names `bionic`, `focal`, etc. are
[Linux development codenames](https://wiki.ubuntu.com/DevelopmentCodeNames)
matching the output of `lsb_release -a`. 
For example, on Ubuntu 18.04:
```shell
$ lsb_release -a
No LSB modules are available.
Distributor ID:	Ubuntu
Description:	Ubuntu 18.04.6 LTS
Release:	18.04
Codename:	bionic
```

On such a platform (`bionic`), the above JSON constructs would be interpreted as:
- Dictionary value:
  ```json
  {
    "a": "A",
    "b": "ubuntu18-B",
    "c": "C"
  }
  ```
- List element:
  ```json
  [
    "A",
    "ubuntu18-B",
    "C"
  ]
  ```

If the current platform's codename is not listed in an `axii.switch_os` block, 
`default` is used as fallback:
- Dictionary value:
  ```json
  {
    "a": "A",
    "b": "B",
    "c": "C"
  }
  ```
- List element:
  ```json
  [
    "A",
    "B",
    "C"
  ]
  ```

#### Nesting

These constructs can also be nested. For example:
```json
{
  "first": {
    "axii.switch_os": {
      "bionic": {
        "second": "ubuntu18"
      },
      "default": {
        "second": {
          "axii.switch_os": {
            "focal": "general/ubuntu20",
            "default": "general"
          }
        }
      }
    }
  }
}
```
are interpreted according to the platform's code name as:
- `bionic`:
  ```json
  {
    "first": {
      "second": "ubuntu18"
    }
  }
  ```
- `focal`:
  ```json
  {
    "first": {
      "second": "general/ubuntu20"
    }
  }
  ```
- `default`:
  ```json
  {
    "first": {
      "second": "general"
    }
  }
  ```

#### Common Use Cases

A few common use case for this syntax are defining OS-specific ...

- Git branches or tags:
  ```json
  {
    "update": {
      "git": {
        "url": "...",
        "default_checkout_ref": {
          "axii.switch_os": {
            "bionic": "ubuntu18",
            "focal": "ubuntu20",
            "jammy": "ubuntu22",
            "buster": "debian10",
            "default": "main"
          }
        }
      }
    }
  }
  ```

- requirements:
  ```json
  {
    "required_modules": {
      "axii.switch_os": {
        "bionic": {
          "apt/gcc-8": {}        
        },
        "focal": {
          "apt/gcc-9": {}        
        },
        "default": {
        }   
      }
    }
  }
  ```

- system packages:
  ```json
  {   
    "system": {
      "packages": {
        "apt": {
          "axii.switch_os": { 
            "focal": [
              "libnlopt-dev",
              "libnlopt-cxx-dev"
            ],
            "default": [
              "libnlopt-dev"
            ]
          }       
        }
      }
    }
  }
  ```


## Common Use Cases and Tricks in Module Definitions

In this section you will find, how the features of module definitions can be used
to reach a certain goal or behavior.


### Override Build Job Number for a Specific Module

You can explicitly override the build job number Axii uses for a specific module by setting
`CMAKE_BUILD_PARALLEL_LEVEL` in the module definition: 

```json
{
  ...,

  "build": {
    "cmake": {
      "env": {
        "CMAKE_BUILD_PARALLEL_LEVEL": "7"
      }
    }
  },

  ...
}
```

This can be helpful if a package is notorious for high memory consumption or not stable when
building with multiple cores.
